import { METADATA_TYPES, PROCESS_STATES, PROCESS_MODES, CLASSIFICATION_STATES } from '@/config'
import { S3FileStatus, GitRefType, ProcessActivityState } from '@/enums'
import { EntityLight } from './entity'
import { ModelVersionLight } from './model'
import { UserLight } from './user'
import { WorkerVersion } from './worker'
import { WorkerConfiguration } from './workerConfiguration'

/**
 * A universally unique identifier, using the format
 * `aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa`.
 */
export type UUID = string

/*
 * Support mutually exclusive object keys.
 * This will still allow the keys to be defined with `undefined` as their value,
 * which is the best we can achieve with TypeScript.
 * From https://stackoverflow.com/q/42123407/#comment123255834_53229567
 */

/**
 * Helper to allow a `keyof` on every key found in any part of a union.
 * While `keyof A | B` will return only the keys found in both `A` and `B`,
 * `UnionKeys<A | B>` returns all the keys in `A` and all the keys in `B`.
 */
export type UnionKeys<T> = T extends T ? keyof T : never

// This does not do much other than copy the type; this just makes autocompletion nicer in all IDEs when using OneOf
type Expand<T> = T extends T ? { [K in keyof T]: T[K] } : never

export type OneOf<T extends object[]> = {
  [K in keyof T]: Expand<T[K] & Partial<Record<Exclude<UnionKeys<T[number]>, keyof T[K]>, never>>>;
}[number]

/**
 * Either all the properties of this type are of their defined type, or all the properties are `null`.
 * There cannot be both null and non-null properties at once, unless some properties in `T` has been declared as `... | null`.
 */
export type NullableProperties<T> = T | { [K in keyof T]: null }

export type Point = [number, number]

export type Polygon = Point[]

export type MetaDataType = keyof typeof METADATA_TYPES

export type ProcessMode = keyof typeof PROCESS_MODES

export type ProcessState = keyof typeof PROCESS_STATES

export type ClassificationState = keyof typeof CLASSIFICATION_STATES

export type DateType = 'exact' | 'lower' | 'upper' | 'unknown'

export type TextOrientation = 'horizontal-lr' | 'horizontal-rl' | 'vertical-lr' | 'vertical-rl'

export type Right = 'read' | 'write' | 'admin'

export interface ImageServer {
  display_name: string
  url: string
  max_width: number | null
  max_height: number | null
}

export interface Image {
  id: UUID
  path: string
  width: number
  height: number
  url: string
  s3_url: string | null
  status: S3FileStatus
  server: ImageServer
}

export interface ZoneLight {
  id: UUID
  polygon: Polygon
}

export interface Zone extends ZoneLight {
  image: Image
  url: string
}

export interface CorpusLight {
  id: UUID
  name: string
  public: boolean
}

export interface ElementType {
  id: UUID
  slug: string
  display_name: string
  folder: boolean
  color: string
}

export interface Corpus extends CorpusLight {
  created: string
  description: string
  indexable: boolean
  authorized_users: number
  rights: Right[]
  types: ElementType[]
  top_level_type: string | null
}

export interface ElementLight {
  id: UUID,
  type: string
  name: string
}

export interface ElementZone extends ElementLight {
  zone: ZoneLight | null
  rotation_angle: number
  mirrored: boolean
}

/**
 * Base interface for any element serializer found in the API.
 * This can be used to build functions that accept any element from anywhere
 */
export interface ElementBase extends ElementLight {
  corpus: CorpusLight | null
  /**
   * URL of the element's thumbnail.
   * Will be null if the element's type is not a folder type.
   * The thumbnail might not have been generated, so 404 errors should be expected.
   */
  thumbnail_url: string | null
  zone: Zone | null
  rotation_angle: number
  mirrored: boolean
  created?: string
  classes?: Classification[] | null
  metadata?: MetaData[] | null
  has_children?: boolean | null
}

export interface ElementTiny extends Required<Pick<ElementBase, 'id' | 'type' | 'name' | 'corpus' | 'thumbnail_url' | 'zone' | 'rotation_angle' | 'mirrored'>> {
  // Corpus is required
  corpus: CorpusLight
}

export interface Element extends ElementTiny {
  created: string
  creator: string | null
  rights: Right[]
  metadata_count: number
  classifications: Classification[]
  worker_version: UUID | null
  worker_run: WorkerRunSummary | null
  confidence: number | null
  /**
   * URL of the element's thumbnail, on which a PUT request can be made to upload a new thumbnail.
   * The URL will only be available if the element's type is a folder type
   * and the user is a corpus admin, or a corpus contributor and the element's creator.
   */
  thumbnail_put_url: string | null
}

export interface ElementList extends Required<Pick<ElementBase, 'id' | 'type' | 'name' | 'corpus' | 'thumbnail_url' | 'zone' | 'rotation_angle' | 'mirrored' | 'created' | 'classes' | 'metadata' | 'has_children'>> {}

export interface PageNumberPagination<T> {
  count: number
  previous: string | null
  next: string | null
  number: number
  results: T[]
}

export interface CursorPagination<T> {
  count: number | null
  previous: string | null
  next: string | null
  results: T[]
}

/**
 * An S3 bucket that can be used for S3 imports.
 */
export interface Bucket {
  name: string
}

export interface WorkerRunSummary {
  id: UUID
  summary: string
}

export interface InterpretedDate {
  type: DateType
  year: number
  month: number | null
  day: number | null
}

export interface AllowedMetaData {
  id: UUID
  type: string
  name: string
}

interface BaseMetaData {
  id: UUID
  type: MetaDataType
  name: string
  value: string | number
  entity_id?: UUID
}

export interface NumericMetaData extends BaseMetaData {
  type: 'numeric'
  value: number
}

export interface StringMetaData extends BaseMetaData {
  type: Exclude<MetaDataType, 'numeric'>
  value: string
}

export type TypedMetaData = StringMetaData | NumericMetaData

export interface MetaDataCreate extends Omit<TypedMetaData, 'id'> {
  elementId: UUID
  worker_version?: UUID
  worker_run_id?: UUID
}

export interface MetaData extends Omit<TypedMetaData, 'entity_id'> {
  dates: InterpretedDate[]
  entity: EntityLight
  worker_version: UUID | null
  worker_run_id: UUID | null
}

export interface Group {
  id: UUID
  name: string
  public: boolean
  members_count: number
  level: number
}

export interface Membership {
  id: UUID
  level: number
  user: UserLight | null
  group: Group | null
}

export interface MLClass {
  id: UUID
  name: string
}

export interface Classification {
  id: UUID
  ml_class: MLClass
  worker_version: UUID | null
  worker_run: WorkerRunSummary | null
  confidence: number | null
  high_confidence: boolean
  state: ClassificationState
}

export interface GitRef {
  id: UUID
  type: GitRefType
  name: string
}

export interface RevisionWithRefs {
  id: UUID
  hash: string
  author: string
  message: string
  created: string
  commit_url: string
  refs: GitRef[]
}

interface ProcessLight {
  id: UUID
  name: string | null
  state: ProcessState
  mode: ProcessMode
  corpus: UUID | null
  activity_state: ProcessActivityState
}

export interface WorkerRun {
  id: UUID
  parents: UUID[]
  worker_version: WorkerVersion
  process: ProcessLight
  configuration: WorkerConfiguration | null
  model_version: ModelVersionLight | null
}

export type JobStatus = 'queued' | 'started' | 'deferred' | 'finished' | 'stopped' | 'scheduled' | 'canceled' | 'failed'

/**
 * An asynchronous job performed by RQ
 */
export interface Job {
  id: UUID
  description: string
  /**
   * Current progress of the job, between 0 and 1. May not be available for every job.
   */
  progress: number | null
  status: JobStatus
  enqueued_at: string | null
  started_at: string | null
  ended_at: string | null
}

/**
 * A cache table listing the origin of results generated on a corpus,
 * based on their worker version, model version and worker configuration.
 *
 * This is called `CorpusWorkerVersion` in the backend.
 */
export interface WorkerVersionCache {
  id: UUID
  worker_version: WorkerVersion
  model_version: ModelVersionLight | null
  worker_configuration: WorkerConfiguration | null
}
