import { assert } from 'chai'
import sinon from 'sinon'
import { parseEntities } from '@/helpers'

describe('helpers/entity', () => {
  let spy

  before('set up console.error spy', () => {
    spy = sinon.spy(console, 'error')
  })

  beforeEach(() => {
    spy.resetHistory()
  })

  after('tear down console.error spy', () => {
    spy.restore()
  })

  const entity1 = {
    id: 'entity1',
    name: 'Vulpix',
    type: 'person',
    worker_version_id: 'versionid'
  }
  const entity2 = {
    id: 'entity2',
    name: 'Ember',
    type: 'misc',
    worker_version_id: null
  }

  it('parses entities', () => {
    assert.deepStrictEqual(
      parseEntities('Vulpix used EMBER!', [
        {
          offset: 0,
          length: 6,
          entity: entity1
        },
        {
          offset: 12,
          length: 5,
          entity: entity2
        }
      ]),
      [
        {
          offset: 0,
          entity: entity1,
          children: [],
          text: 'Vulpix',
          overflow: false
        },
        {
          offset: 6,
          text: ' used ',
          children: []
        },
        {
          offset: 12,
          entity: entity2,
          children: [],
          text: 'EMBER',
          overflow: false
        },
        {
          offset: 17,
          text: '!',
          children: []
        }
      ]
    )
    assert.ok(spy.notCalled)
  })

  it('handles no entities at all', () => {
    assert.deepStrictEqual(
      parseEntities('Lorem ipsum dolor sit amet', []),
      [
        {
          offset: 0,
          text: 'Lorem ipsum dolor sit amet',
          children: []
        }
      ]
    )
    assert.ok(spy.notCalled)
  })

  it('ignores entities with negative offsets', () => {
    assert.deepStrictEqual(
      parseEntities('Vulpix used EMBER!', [
        {
          offset: -42,
          length: 1,
          entity: entity1
        },
        {
          offset: 0,
          length: 6,
          entity: entity1
        }
      ]),
      [
        {
          offset: 0,
          text: 'Vulpix',
          entity: entity1,
          children: [],
          overflow: false
        },
        {
          offset: 6,
          text: ' used EMBER!',
          children: []
        }
      ]
    )
    assert.ok(spy.calledOnceWithExactly('Entity entity1 has a negative offset.'))
  })

  it('warns on entities longer than the text', () => {
    assert.deepStrictEqual(
      parseEntities('Vulpix', [
        {
          offset: 0,
          length: 9001,
          entity: entity1
        }
      ]),
      [
        {
          offset: 0,
          text: 'Vulpix',
          entity: entity1,
          children: [],
          overflow: true
        }
      ]
    )
    assert.ok(spy.notCalled)
  })

  it('supports parsing a portion of text', () => {
    assert.deepStrictEqual(
      parseEntities('used EMBER!', [
        {
          offset: 12,
          length: 5,
          entity: entity2
        }
      ], 7),
      [
        {
          offset: 7,
          text: 'used ',
          children: []
        },
        {
          offset: 12,
          entity: entity2,
          children: [],
          text: 'EMBER',
          overflow: false
        },
        {
          offset: 17,
          text: '!',
          children: []
        }
      ]
    )
    assert.ok(spy.notCalled)
  })

  it('supports nested entities', () => {
    assert.deepStrictEqual(
      parseEntities('Vulpemberix', [
        {
          offset: 0,
          length: 11,
          entity: entity1
        },
        {
          offset: 4,
          length: 5,
          entity: entity2
        }
      ]),
      [
        {
          offset: 0,
          entity: entity1,
          text: 'Vulpemberix',
          children: [
            {
              offset: 4,
              length: 5,
              entity: entity2
            }
          ],
          overflow: false
        }
      ]
    )
    assert.ok(spy.notCalled)
  })

  it('parses overlapping entities', () => {
    assert.deepStrictEqual(
      parseEntities('Vulpix used EMBER!', [
        {
          offset: 0,
          // "Vulpix used"
          length: 11,
          entity: entity1
        },
        {
          offset: 7,
          // "used EMBER"
          length: 10,
          entity: entity2
        }
      ]),
      [
        {
          offset: 0,
          entity: entity1,
          text: 'Vulpix used',
          children: [
            {
              offset: 7,
              /*
               * The length exceeds "Vulpix used", so we will get a warning about an entity exceeding the text
               * once the component calls parseEntities on the children.
               */
              length: 10,
              entity: entity2
            }
          ],
          overflow: false
        },
        {
          offset: 11,
          text: ' EMBER!',
          children: []
        }
      ]
    )
    assert.ok(spy.notCalled)
  })
})
