import { assert } from 'chai'
import { iiifUri } from '@/helpers'

describe('helpers/iiif', () => {
  describe('iiifUri', () => {
    const image = {
      url: 'http://lol/1234/',
      width: 500,
      height: 700
    }
    const zone = {
      image,
      polygon: [[100, 200], [100, 600], [400, 600], [400, 200], [100, 200]],
      url: 'http://not/this/url'
    }

    it('requires an image in the zone', () => {
      assert.throws(
        () => iiifUri({ polygon: zone.polygon }),
        Error,
        'An image is required.'
      )
    })

    it('requires an image with valid dimensions', () => {
      assert.throws(
        () => iiifUri({
          image: { url: 'http://blah' },
          polygon: zone.polygon
        }),
        Error,
        'An image with valid dimensions is required.'
      )
    })

    it('requires an image with a URL', () => {
      assert.throws(
        () => iiifUri({
          image: { width: 42, height: 42 },
          polygon: zone.polygon
        }),
        Error,
        'An image with a valid URL is required.'
      )
    })

    it('crops using the polygon', () => {
      assert.strictEqual(
        iiifUri(zone),
        'http://lol/1234/100,200,300,400/full/0/default.jpg'
      )
    })

    it('does not crop when possible', () => {
      assert.strictEqual(
        iiifUri({ image, polygon: [[-50, -50], [9000, 9000]] }),
        'http://lol/1234/full/full/0/default.jpg'
      )
    })

    it('does not crop without a polygon', () => {
      assert.strictEqual(
        iiifUri({ image }),
        'http://lol/1234/full/full/0/default.jpg'
      )
    })

    it('can resize by width only', () => {
      assert.strictEqual(
        iiifUri(zone, { width: 50 }),
        'http://lol/1234/100,200,300,400/50,/0/default.jpg'
      )
    })

    it('can resize by height only', () => {
      assert.strictEqual(
        iiifUri(zone, { height: 50 }),
        'http://lol/1234/100,200,300,400/,50/0/default.jpg'
      )
    })

    it('can resize with both width and height', () => {
      assert.strictEqual(
        iiifUri(zone, { width: 50, height: 50 }),
        'http://lol/1234/100,200,300,400/!50,50/0/default.jpg'
      )
    })

    it('ignores a maximum width above the bounding box width', () => {
      assert.strictEqual(
        iiifUri(zone, { width: 301 }),
        'http://lol/1234/100,200,300,400/full/0/default.jpg'
      )
    })

    it('ignores a maximum height above the bounding box height', () => {
      assert.strictEqual(
        iiifUri(zone, { height: 401 }),
        'http://lol/1234/100,200,300,400/full/0/default.jpg'
      )
    })

    it('prevents resizing by width over 100%', () => {
      assert.strictEqual(
        iiifUri(zone, { width: 600, height: 200 }),
        'http://lol/1234/100,200,300,400/!300,100/0/default.jpg'
      )
    })

    it('prevents resizing by height over 100%', () => {
      assert.strictEqual(
        iiifUri(zone, { width: 200, height: 800 }),
        'http://lol/1234/100,200,300,400/!100,400/0/default.jpg'
      )
    })

    it('can add margins to a crop', () => {
      assert.strictEqual(
        iiifUri(zone, { margin: 10 }),
        'http://lol/1234/90,190,320,420/full/0/default.jpg'
      )
    })

    it('ignores a maximum width above the bounding box width with margins', () => {
      assert.strictEqual(
        iiifUri(zone, { margin: 10, width: 321 }),
        'http://lol/1234/90,190,320,420/full/0/default.jpg'
      )
    })

    it('ignores a maximum height above the bounding box height with margins', () => {
      assert.strictEqual(
        iiifUri(zone, { margin: 10, height: 421 }),
        'http://lol/1234/90,190,320,420/full/0/default.jpg'
      )
    })

    it('prevents resizing by width over 100% with margins', () => {
      assert.strictEqual(
        iiifUri(zone, { margin: 10, width: 640, height: 210 }),
        'http://lol/1234/90,190,320,420/!320,105/0/default.jpg'
      )
    })

    it('prevents resizing by height over 100% with margins', () => {
      assert.strictEqual(
        iiifUri(zone, { margin: 10, width: 200, height: 840 }),
        'http://lol/1234/90,190,320,420/!100,420/0/default.jpg'
      )
    })
  })
})
