import { assert } from 'chai'
import { setActivePinia } from 'pinia'
import axios from 'axios'
import { useElementsStore, useNotificationStore } from '@/stores'
import { assertRejects, FakeAxios, setUpTestPinia, actionsCompletedPlugin } from '../testhelpers.js'
import { elementDatasetsSample } from '../samples.js'

describe('elements.ts', () => {
  let pinia, mock, store, notificationStore

  before(() => {
    [pinia] = setUpTestPinia([actionsCompletedPlugin])
    setActivePinia(pinia)
    mock = new FakeAxios(axios)
    store = useElementsStore()
    notificationStore = useNotificationStore()
  })

  beforeEach(() => {
    mock.reset()
    store.$reset()
    notificationStore.$reset()
  })

  after('Removing Axios mock', () => {
    mock.restore()
  })

  describe('actions', () => {
    describe('listElementDatasets', () => {
      it('returns early in case datasets have been fetched', async () => {
        store.elementDatasets = { element_id: [] }
        await store.listElementDatasets('element_id')
        await store.actionsCompleted()
      })

      it('automatically loads datasets for an element', async () => {
        const pagination1 = {
          count: 2,
          previous: null,
          next: '/element/element_id/datasets/?page=2',
          number: 1,
          results: [elementDatasetsSample.results[0]]
        }
        const pagination2 = {
          count: 2,
          previous: 'http://previous',
          next: null,
          number: 2,
          results: [elementDatasetsSample.results[1]]
        }
        mock.onGet('/element/element_id/datasets/', { params: {} }).reply(200, pagination1)
        mock.onGet('/element/element_id/datasets/', { params: { page: 2 } }).reply(200, pagination2)

        await store.listElementDatasets('element_id')
        await store.actionsCompleted()

        assert.deepStrictEqual(store.elementDatasets, {
          element_id: [
            elementDatasetsSample.results[0],
            elementDatasetsSample.results[1]
          ]
        })
      })

      it('handles errors', async () => {
        mock.onGet('/element/element_id/datasets/').reply(403, { detail: ['Forbidden'] })

        await assertRejects(async () =>
          await store.listElementDatasets('element_id')
        )
        assert.deepStrictEqual(notificationStore.notifications, [
          { id: 0, type: 'error', text: 'Forbidden' }
        ])
      })
    })
  })
})
