import axios from 'axios'
import { assert } from 'chai'
import { pick } from 'lodash'
import { setActivePinia, createPinia } from 'pinia'
import { useWorkerStore, useRepositoryStore } from '@/stores'
import { reposSample, repoSample } from '../samples'
import { FakeAxios } from '../testhelpers'

describe('repos', () => {
  describe('actions', () => {
    let mock, store, workerStore

    before('Setting up Axios mock', () => {
      mock = new FakeAxios(axios)
      setActivePinia(createPinia())
      store = useRepositoryStore()
      workerStore = useWorkerStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.$reset()
      workerStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('listRepository', () => {
      it('lists existing repositories', async () => {
        const sample = {
          ...reposSample,
          count: 2,
          results: [
            ...reposSample.results,
            {
              id: 'other_repo',
              workers: [
                {
                  id: 'worker 2'
                }, {
                  id: 'worker 3'
                }
              ]
            }
          ]
        }

        mock.onGet('/process/repos/').reply(200, sample)

        const page = await store.listRepository()
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/process/repos/'
          }
        ])
        assert.deepStrictEqual(page, sample)
        assert.deepStrictEqual(Object.keys(workerStore.workers), ['worker 1', 'worker 2', 'worker 3'])
      })
    })

    describe('deleteRepository', () => {
      it('deletes a repo', async () => {
        mock.onDelete('/process/repos/repoid/').reply(204)

        store.repositories = { repoid: {}, secondId: {} }
        await store.deleteRepository('repoid')

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'delete',
            url: '/process/repos/repoid/'
          }
        ])
      })
    })

    describe('retrieveRepository', () => {
      it('retrieves a repo', async () => {
        mock.onGet('/process/repos/repoid/').reply(200, repoSample)
        await store.retrieveRepository('repoid')
        assert.deepStrictEqual(store.repositories, {
          repoid: repoSample
        })
      })
    })
  })
})
